import pytest
from dbt.tests.adapter.utils.base_utils import BaseUtils
from dbt.tests.adapter.utils.fixture_get_powers_of_two import (
    models__test_get_powers_of_two_sql,
    models__test_get_powers_of_two_yml,
)


class BaseGetPowersOfTwo(BaseUtils):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_get_powers_of_two.yml": models__test_get_powers_of_two_yml,
            "test_get_powers_of_two.sql": self.interpolate_macro_namespace(
                models__test_get_powers_of_two_sql, "get_powers_of_two"
            ),
        }


class TestGetPowersOfTwo(BaseGetPowersOfTwo):
    pass



models__sql_header = """
{% call set_sql_header(config) %}
with _variables as (
    select 1 as my_variable
)
{%- endcall %}
select my_variable from _variables
"""

models__ephemeral_model = """
{{ config(materialized = 'ephemeral') }}
select
    coalesce(sample_num, 0) + 10 as col_deci
from {{ ref('sample_model') }}
"""

models__second_ephemeral_model = """
{{ config(materialized = 'ephemeral') }}
select
    col_deci + 100 as col_hundo
from {{ ref('ephemeral_model') }}
"""

models__sample_model = """
select * from {{ ref('sample_seed') }}
"""

seeds__sample_seed = """sample_num,sample_bool
1,true
2,false
3,true
4,false
5,true
6,false
7,true
"""



# If generate_series works properly, there should be no `null` values in the resulting model

models__test_generate_series_sql = """
with generated_numbers as (
    {{ dbt.generate_series(10) }}
), expected_numbers as (
    select 1 as expected
    union all
    select 2 as expected
    union all
    select 3 as expected
    union all
    select 4 as expected
    union all
    select 5 as expected
    union all
    select 6 as expected
    union all
    select 7 as expected
    union all
    select 8 as expected
    union all
    select 9 as expected
    union all
    select 10 as expected
), joined as (
    select
        generated_numbers.generated_number,
        expected_numbers.expected
    from generated_numbers
    left join expected_numbers on generated_numbers.generated_number = expected_numbers.expected
)

SELECT * from joined
"""

models__test_generate_series_yml = """
version: 2
models:
  - name: test_generate_series
    tests:
      - assert_equal:
          actual: generated_number
          expected: expected
"""



from typing import Type

import pytest

from dbt.adapters.base.impl import BaseAdapter
from dbt.exceptions import DbtRuntimeError, InvalidConnectionError


class BaseValidateSqlMethod:
    """Tests the behavior of the validate_sql method for the relevant adapters.

    The valid and invalid SQL should work with most engines by default, but
    both inputs can be overridden as needed for a given engine to get the correct
    behavior.

    The base method is meant to throw the appropriate custom exception when validate_sql
    fails.
    """

    @pytest.fixture(scope="class")
    def valid_sql(self) -> str:
        """Returns a valid statement for issuing as a validate_sql query.

        Ideally this would be checkable for non-execution. For example, we could use a
        CREATE TABLE statement with an assertion that no table was created. However,
        for most adapter types this is unnecessary - the EXPLAIN keyword has exactly the
        behavior we want, and here we are essentially testing to make sure it is
        supported. As such, we return a simple SELECT query, and leave it to
        engine-specific test overrides to specify more detailed behavior as appropriate.
        """

        return "select 1"

    @pytest.fixture(scope="class")
    def invalid_sql(self) -> str:
        """Returns an invalid statement for issuing a bad validate_sql query."""

        return "Let's run some invalid SQL and see if we get an error!"

    @pytest.fixture(scope="class")
    def expected_exception(self) -> Type[Exception]:
        """Returns the Exception type thrown by a failed query.

        Defaults to dbt.exceptions.DbtRuntimeError because that is the most common
        base exception for adapters to throw."""
        return DbtRuntimeError

    def test_validate_sql_success(self, adapter: BaseAdapter, valid_sql: str) -> None:
        """Executes validate_sql on valid SQL. No news is good news."""
        with adapter.connection_named("test_valid_sql_validation"):
            adapter.validate_sql(valid_sql)

    def test_validate_sql_failure(
        self,
        adapter: BaseAdapter,
        invalid_sql: str,
        expected_exception: Type[Exception],
    ) -> None:
        """Executes validate_sql on invalid SQL, expecting the exception."""
        with pytest.raises(expected_exception=expected_exception) as excinfo:
            with adapter.connection_named("test_invalid_sql_validation"):
                adapter.validate_sql(invalid_sql)

        # InvalidConnectionError is a subclass of DbtRuntimeError, so we have to handle
        # it separately.
        if excinfo.type == InvalidConnectionError:
            raise ValueError(
                "Unexpected InvalidConnectionError. This typically indicates a problem "
                "with the test setup, rather than the expected error for an invalid "
                "validate_sql query."
            ) from excinfo.value


class TestValidateSqlMethod(BaseValidateSqlMethod):
    pass



import pytest
from dbt.tests.util import run_dbt, check_relations_equal
from collections import namedtuple


models__merge_exclude_columns_sql = """
{{ config(
    materialized = 'incremental',
    unique_key = 'id',
    incremental_strategy='merge',
    merge_exclude_columns=['msg']
) }}

{% if not is_incremental() %}

-- data for first invocation of model

select 1 as id, 'hello' as msg, 'blue' as color
union all
select 2 as id, 'goodbye' as msg, 'red' as color

{% else %}

-- data for subsequent incremental update

select 1 as id, 'hey' as msg, 'blue' as color
union all
select 2 as id, 'yo' as msg, 'green' as color
union all
select 3 as id, 'anyway' as msg, 'purple' as color

{% endif %}
"""

seeds__expected_merge_exclude_columns_csv = """id,msg,color
1,hello,blue
2,goodbye,green
3,anyway,purple
"""

ResultHolder = namedtuple(
    "ResultHolder",
    [
        "seed_count",
        "model_count",
        "seed_rows",
        "inc_test_model_count",
        "relation",
    ],
)


class BaseMergeExcludeColumns:
    @pytest.fixture(scope="class")
    def models(self):
        return {"merge_exclude_columns.sql": models__merge_exclude_columns_sql}

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"expected_merge_exclude_columns.csv": seeds__expected_merge_exclude_columns_csv}

    def update_incremental_model(self, incremental_model):
        """update incremental model after the seed table has been updated"""
        model_result_set = run_dbt(["run", "--select", incremental_model])
        return len(model_result_set)

    def get_test_fields(self, project, seed, incremental_model, update_sql_file):

        seed_count = len(run_dbt(["seed", "--select", seed, "--full-refresh"]))

        model_count = len(run_dbt(["run", "--select", incremental_model, "--full-refresh"]))

        relation = incremental_model
        # update seed in anticipation of incremental model update
        row_count_query = "select * from {}.{}".format(project.test_schema, seed)

        seed_rows = len(project.run_sql(row_count_query, fetch="all"))

        # propagate seed state to incremental model according to unique keys
        inc_test_model_count = self.update_incremental_model(incremental_model=incremental_model)

        return ResultHolder(seed_count, model_count, seed_rows, inc_test_model_count, relation)

    def check_scenario_correctness(self, expected_fields, test_case_fields, project):
        """Invoke assertions to verify correct build functionality"""
        # 1. test seed(s) should build afresh
        assert expected_fields.seed_count == test_case_fields.seed_count
        # 2. test model(s) should build afresh
        assert expected_fields.model_count == test_case_fields.model_count
        # 3. seeds should have intended row counts post update
        assert expected_fields.seed_rows == test_case_fields.seed_rows
        # 4. incremental test model(s) should be updated
        assert expected_fields.inc_test_model_count == test_case_fields.inc_test_model_count
        # 5. result table should match intended result set (itself a relation)
        check_relations_equal(
            project.adapter, [expected_fields.relation, test_case_fields.relation]
        )

    def test__merge_exclude_columns(self, project):
        """seed should match model after two incremental runs"""

        expected_fields = ResultHolder(
            seed_count=1,
            model_count=1,
            inc_test_model_count=1,
            seed_rows=3,
            relation="expected_merge_exclude_columns",
        )

        test_case_fields = self.get_test_fields(
            project,
            seed="expected_merge_exclude_columns",
            incremental_model="merge_exclude_columns",
            update_sql_file=None,
        )
        self.check_scenario_correctness(expected_fields, test_case_fields, project)



# get_powers_of_two

models__test_get_powers_of_two_sql = """
select {{ get_powers_of_two(1) }} as actual, 1 as expected

union all

select {{ get_powers_of_two(4) }} as actual, 2 as expected

union all

select {{ get_powers_of_two(27) }} as actual, 5 as expected

union all

select {{ get_powers_of_two(256) }} as actual, 8 as expected

union all

select {{ get_powers_of_two(3125) }} as actual, 12 as expected

union all

select {{ get_powers_of_two(46656) }} as actual, 16 as expected

union all

select {{ get_powers_of_two(823543) }} as actual, 20 as expected
"""

models__test_get_powers_of_two_yml = """
version: 2
models:
  - name: test_powers_of_two
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""



import pytest
import json
from dbt.exceptions import DbtRuntimeError
from dbt.version import __version__ as dbt_version
from dbt.tests.util import run_dbt_and_capture
from dbt.tests.adapter.query_comment.fixtures import MACROS__MACRO_SQL, MODELS__X_SQL


class BaseDefaultQueryComments:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "x.sql": MODELS__X_SQL,
        }

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "macro.sql": MACROS__MACRO_SQL,
        }

    def run_get_json(self, expect_pass=True):
        res, raw_logs = run_dbt_and_capture(
            ["--debug", "--log-format=json", "run"], expect_pass=expect_pass
        )

        # empty lists evaluate as False
        assert len(res) > 0
        return raw_logs


# Base setup to be inherited #
class BaseQueryComments(BaseDefaultQueryComments):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"query-comment": "dbt\nrules!\n"}

    def test_matches_comment(self, project) -> bool:
        logs = self.run_get_json()
        assert r"/* dbt\nrules! */\n" in logs


class BaseMacroQueryComments(BaseDefaultQueryComments):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"query-comment": "{{ query_header_no_args() }}"}

    def test_matches_comment(self, project) -> bool:
        logs = self.run_get_json()
        assert r"/* dbt macros\nare pretty cool */\n" in logs


class BaseMacroArgsQueryComments(BaseDefaultQueryComments):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"query-comment": "{{ return(ordered_to_json(query_header_args(target.name))) }}"}

    def test_matches_comment(self, project) -> bool:
        logs = self.run_get_json()
        expected_dct = {
            "app": "dbt++",
            "dbt_version": dbt_version,
            "macro_version": "0.1.0",
            "message": f"blah: {project.adapter.config.target_name}",
        }
        expected = r"/* {} */\n".format(json.dumps(expected_dct, sort_keys=True)).replace(
            '"', r"\""
        )
        assert expected in logs


class BaseMacroInvalidQueryComments(BaseDefaultQueryComments):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"query-comment": "{{ invalid_query_header() }}"}

    def test_run_assert_comments(self, project):
        with pytest.raises(DbtRuntimeError):
            self.run_get_json(expect_pass=False)


class BaseNullQueryComments(BaseDefaultQueryComments):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"query-comment": None}

    def test_matches_comment(self, project) -> bool:
        logs = self.run_get_json()
        assert "/*" not in logs or "*/" not in logs


class BaseEmptyQueryComments(BaseDefaultQueryComments):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"query-comment": ""}

    def test_matches_comment(self, project) -> bool:
        logs = self.run_get_json()
        assert "/*" not in logs or "*/" not in logs


# Tests #
class TestQueryComments(BaseQueryComments):
    pass


class TestMacroQueryComments(BaseMacroQueryComments):
    pass


class TestMacroArgsQueryComments(BaseMacroArgsQueryComments):
    pass


class TestMacroInvalidQueryComments(BaseMacroInvalidQueryComments):
    pass


class TestNullQueryComments(BaseNullQueryComments):
    pass


class TestEmptyQueryComments(BaseEmptyQueryComments):
    pass
